﻿using System;
using System.Collections.Generic;
using System.Linq;
using VDS.RDF;
using VDS.RDF.Query;
using VDS.RDF.Update;
using VDS.RDF.Storage;
using VDS.RDF.Writing;
using System.IO;
using VDS.RDF.Shacl;

namespace Coscine.Metadata
{
    public class RdfStoreConnector: IDisposable
    {
        public readonly SparqlRemoteUpdateEndpoint UpdateEndpoint;
        public readonly SparqlRemoteEndpoint QueryEndpoint;
        public readonly ReadWriteSparqlConnector ReadWriteSparqlConnector;
        public readonly string UserUrlPrefix = "https://purl.org/coscine/users";
        public readonly Uri OrgPrefixUrl = new Uri("http://www.w3.org/ns/org#");
        public readonly Uri FoafPrefixUrl = new Uri("http://xmlns.com/foaf/0.1/");

        public RdfStoreConnector(string sparqlEndpoint = "http://localhost:8890/sparql")
        {
            UpdateEndpoint = new SparqlRemoteUpdateEndpoint(new Uri(string.Format(sparqlEndpoint)));
            QueryEndpoint = new SparqlRemoteEndpoint(new Uri(string.Format(sparqlEndpoint)));
            ReadWriteSparqlConnector = new ReadWriteSparqlConnector(QueryEndpoint, UpdateEndpoint);
        }

        public bool HasGraph(string graphName)
        {
            return HasGraph(new Uri(graphName));
        }

        public bool HasGraph(Uri graphUri)
        {
            return !GetGraph(graphUri).IsEmpty;
        }

        public IEnumerable<Uri> ListGraphs()  
        {
            return ReadWriteSparqlConnector.ListGraphs();
        }
        
        public IEnumerable<Uri> GetApplicationProfiles()
        {
            var commandString = new SparqlParameterizedString
            {
                CommandText = @"SELECT DISTINCT ?s
                                WHERE
                                {
                                    ?s a <http://www.w3.org/ns/shacl#NodeShape>
                                }"
            };
            var resultSet = QueryEndpoint.QueryWithResultSet(commandString.ToString());
            return resultSet.Select(x => new Uri(x.Value("s").ToString()));
        }

        // Returns an empty graph, if the graph does not exists
        public IGraph GetGraph(string graphName)
        {
            return GetGraph(new Uri(graphName));
        }

        // Returns an empty graph, if the graph does not exists
        public IGraph GetGraph(Uri graphUri)
        {
            var graph = new Graph();

            ReadWriteSparqlConnector.LoadGraph(graph, graphUri.AbsoluteUri);

            return graph;
        }

        public void AddGraph(IGraph graph)
        {
            ReadWriteSparqlConnector.SaveGraph(graph);
        }

        public IGraph ParseGraph(string data, IRdfReader parser)
        {
            var graph = new Graph();
            parser.Load(graph, new StringReader(data));
            return graph;
        }

        public string SerializeGraph(IGraph graph, BaseRdfWriter writer)
        {
            return VDS.RDF.Writing.StringWriter.Write(graph, writer);
        }

        public void ClearGraph(string graphName)
        {
            ClearGraph(new Uri(graphName));
        }

        public void DeleteGraph(string graphName)
        {
            DeleteGraph(new Uri(graphName));
        }

        public void CreateNamedGraph(string graphName)
        {
            CreateNamedGraph(new Uri(graphName));
        }

        public void ClearGraph(Uri graphUri)
        {
            SparqlParameterizedString queryString = new SparqlParameterizedString
            {
                CommandText = "CLEAR GRAPH @graph"
            };
            queryString.SetUri("graph", graphUri);
            QueryEndpoint.QueryRaw(queryString.ToString());
        }

        public void DeleteGraph(Uri graphUri)
        {
            ReadWriteSparqlConnector.DeleteGraph(graphUri.AbsoluteUri);
        }

        public void CreateNamedGraph(Uri graphUri)
        {
            SparqlParameterizedString queryString = new SparqlParameterizedString
            {
                CommandText = "CREATE GRAPH @graph"
            };
            queryString.SetUri("graph", graphUri);
            QueryEndpoint.QueryRaw(queryString.ToString());
        }

        public void ParseAndAddGraph(Uri graphUri, string data, IRdfReader parser)
        {
            if (HasGraph(graphUri))
            {
                ClearGraph(graphUri);
            }
            else
            {
                CreateNamedGraph(graphUri);
            }

            var graph = ParseGraph(data, parser);

            AddGraph(graph);
        }

        public void ParseAndAddGraph(string data, IRdfReader parser)
        {
            var graph = ParseGraph(data, parser);
            var graphName = graph.BaseUri.ToString();

            if (HasGraph(graphName))
            {
                ClearGraph(graphName);
            }
            else
            {
                CreateNamedGraph(graphName);
            }

            AddGraph(graph);
        }

        private Triple GetDefinitionTriple(IGraph graph, Uri graphName)
        {
            var triples = graph.GetTriplesWithSubjectPredicate(graph.CreateUriNode(graphName), graph.CreateUriNode(new Uri("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")));
            if (triples.Count() == 1)
            {
                return triples.First();
            }
            else
            {
                throw new Exception($"The input must specify exactly 1 schema. Your input has {triples.Count()} schemas.");
            }
        }

        public Uri GetApplicationProfileUri(IGraph graph, Uri graphName)
        {
            var triple = GetDefinitionTriple(graph, graphName);
            return FormatApplicationProfileUrl(triple);
        }

        public IEnumerable<Uri> GetVocabularies(IGraph graph)
        {
            return graph.GetTriplesWithPredicate(new Uri("http://www.w3.org/ns/shacl#class")).Where(x => x.Object.NodeType == NodeType.Uri).Select(x => new Uri(x.Object.ToString())).Distinct();
        }

        public IEnumerable<Uri> GetClasses(IGraph graph)
        {
            return graph.GetTriplesWithPredicate(new Uri("http://www.w3.org/ns/shacl#class")).Where(x => x.Object.NodeType == NodeType.Uri).Select(x => new Uri(x.Object.ToString())).Distinct();
        }

        public bool ValidateShacl(IGraph graph, Uri graphName)
        {
            // Create a copy of the input graph
            var dataGraph = new Graph();

            dataGraph.Merge(graph);

            AdaptDataGraph(dataGraph, graphName);

            var shapesGraph = GetGraph(GetApplicationProfileUri(dataGraph, graphName));

            foreach (var classElement in GetClasses(shapesGraph))
            {
                var classGraph = GetClassGraph(classElement);
                dataGraph.Merge(classGraph);
            }

            var processor = new ShapesGraph(shapesGraph);
            var conforms = processor.Conforms(dataGraph);

            return conforms;
        }

        public IGraph GetClassGraph(Uri classElement)
        {
            // construct graph with all instances of classElement
            SparqlParameterizedString cmdString = new SparqlParameterizedString
            {
                CommandText = "CONSTRUCT { ?s a @classElement . ?s ?p ?o } WHERE { ?s a ?class . ?class <http://www.w3.org/2000/01/rdf-schema#subClassOf>* @classElement . ?s ?p ?o }"
            };
            cmdString.SetUri("classElement", classElement);

            var resultGraph = QueryEndpoint.QueryWithResultGraph(cmdString.ToString());
            return resultGraph;
        }

        private void AdaptDataGraph(Graph graph, Uri graphName)
        {
            var triple = GetDefinitionTriple(graph, graphName);
            graph.Retract(triple);
            graph.Assert(triple.Subject, triple.Predicate, graph.CreateUriNode(FormatApplicationProfileUrl(triple)));
        }

        private static Uri FormatApplicationProfileUrl(Triple triple)
        {
            var objectString = triple.Object.ToString();
            if (objectString[objectString.Length - 1] != '/')
            {
                objectString += '/';
            }

            return new Uri(objectString);
        }

        public Dictionary<string, string> GetVocabularyLabels(IGraph graph, string langSuffix)
        {
            var triples = graph.GetTriplesWithPredicate(new Uri("http://www.w3.org/2000/01/rdf-schema#label"));
            var dict = triples.Where(x => x.Object.NodeType == NodeType.Literal)
                .Where(x => ((LiteralNode)x.Object).Language == langSuffix)
                .ToDictionary(triple => triple.Subject.ToString(), triple => ((LiteralNode)triple.Object).Value);
            return dict;
        }

        public IEnumerable<Triple> GetTriplesByObject(Uri subject, Uri predicate, Uri obj)
        {
            return GetTriplesByObjectWrapper(subject, predicate, (cmdString) => cmdString.SetUri("o", obj));
        }

        public IEnumerable<Triple> GetTriplesByObject(Uri subject, Uri predicate, string obj)
        {
            return GetTriplesByObjectWrapper(subject, predicate, (cmdString) => cmdString.SetLiteral("o", obj));
        }

        private IEnumerable<Triple> GetTriplesByObjectWrapper(Uri subject, Uri predicate, Action<SparqlParameterizedString> objWrapper)
        {
            SparqlParameterizedString cmdString = new SparqlParameterizedString
            {
                CommandText = "SELECT ?s ?p ?o WHERE { ?s ?p ?o . " +
                ((subject != null) ? "@s " : "?s ") +
                ((predicate != null) ? "@p " : "?p ") +
                "@o }"
            };
            if (subject != null)
            {
                cmdString.SetUri("s", subject);
            }
            if (predicate != null)
            {
                cmdString.SetUri("p", predicate);
            }
            objWrapper(cmdString);

            var resultSet = QueryEndpoint.QueryWithResultSet(cmdString.ToString());

            var triples = new List<Triple>();
            foreach (SparqlResult r in resultSet)
            {
                triples.Add(new Triple(r.Value("s"), r.Value("p"), r.Value("o")));
            }
            return triples;
        }

        public IEnumerable<Triple> GetLabelForSubject(Uri subject)
        {
            SparqlParameterizedString cmdString = new SparqlParameterizedString
            {
                CommandText = "PREFIX rdfs: < http://www.w3.org/2000/01/rdf-schema#> " +
                        "SELECT DISTINCT (@baseGraph AS ?s) (rdfs:label AS ?p) ?o " +
                        "WHERE { " +
                        "@baseGraph rdfs:label ?o . }"
            };
            cmdString.SetUri("baseGraph", subject);
            var resultSet = QueryEndpoint.QueryWithResultSet(cmdString.ToString());
            var triples = new List<Triple>();
            foreach (SparqlResult r in resultSet)
            {
                triples.Add(new Triple(r.Value("s"), r.Value("p"), r.Value("o")));
            }
            return triples;
        }

        public IEnumerable<Triple> GetTriples(Uri graph, Uri obj, string searchterm = null, int member = 0, List<string> externalIds = null, int numberOfResults = 100, int offset = 0)
        {
            var graphFilter = graph != null ? "FROM @g " : "";
            var objectFilter = obj != null ? "?s ?p @o . " : "";
            var searchFilter = (searchterm != null) ? "filter contains(LCASE(?o), LCASE(@searchterm)) . " : "";
            var membersAndExternalIdFilter = (member == 1 && externalIds != null) ?
                @"{ 
                    SELECT DISTINCT ?s WHERE 
                        { ?class ?p ?memberUrl  
                            ; org:organization ?s . 
                            { SELECT DISTINCT ?memberUrl WHERE { 
                                ?memberUrl ?p ?value . FILTER(?value IN ( " + PrepareIncludeString(externalIds) + @" ))
                            } } 
                    } 
                }"
                : "";

            SparqlParameterizedString cmdString = new SparqlParameterizedString
            {
                CommandText = 
                    "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> " +
                    "PREFIX foaf: <http://xmlns.com/foaf/0.1/> " +
                    "PREFIX org: <http://www.w3.org/ns/org#> " +

                    "SELECT DISTINCT ?s (rdfs:label AS ?p) ?o " +
                    graphFilter +
                    "WHERE { " +
                        objectFilter +
                        "?s rdfs:label ?o . " +
                        searchFilter +
                        membersAndExternalIdFilter +
                    "} " +
                    "LIMIT @numberOfResults " +
                    "OFFSET @offset "
            };

            // bind all the required values
            if (graph != null)
            {
                cmdString.SetUri("g", graph);
            }
            if (obj != null)
            {
                cmdString.SetUri("o", obj);
            }
            if (searchterm != null)
            {
                cmdString.SetLiteral("searchterm", searchterm);
            }
            if (externalIds != null)
            {
                for (var i = 0; i < externalIds.Count; i++)
                {
                    cmdString.SetLiteral("externalId" + i, externalIds.ElementAt(i));
                }
            }
            cmdString.SetLiteral("numberOfResults", numberOfResults);
            cmdString.SetLiteral("offset", offset);

            var resultSet = QueryEndpoint.QueryWithResultSet(cmdString.ToString());
            var baseUrls = new HashSet<Uri>();
            var triples = new List<Triple>();

            foreach (SparqlResult r in resultSet)
            {
                var baseUrl = r.Value("s").ToString();
                if (member == 1 && baseUrl.IndexOf("#") != -1)
                {
                    baseUrls.Add(new Uri(baseUrl.Substring(0, baseUrl.LastIndexOf("#"))));
                }
                triples.Add(new Triple(r.Value("s"), r.Value("p"), r.Value("o")));
            }

            // add the top level organization
            foreach(var url in baseUrls)
            {
                triples.AddRange(GetLabelForSubject(url));
            }

            return triples;
        }

        private string PrepareIncludeString(List<string> externalIds)
        {
            var res = "";
            for (var i = 0; i < externalIds.Count; i++)
            {
                res += "@externalId" + i;
                if (i != externalIds.Count - 1)
                {
                    res += " , ";
                }
            }
            return res;
        }

        public void Dispose()
        {
            ReadWriteSparqlConnector.Dispose();
        }

        public string GetIkzForOrganization(Uri subject)
        {
            SparqlParameterizedString cmdString = new SparqlParameterizedString
            {
                CommandText = "SELECT SUBSTR( ?o, 5 ) AS ?o WHERE {{GRAPH < https://ror.org/04xfq0f34> { @subject <http://www.w3.org/ns/org#identifier> ?o .FILTER strStarts( ?o, \"ikz:\" )}}}}"
            };
            cmdString.SetUri("baseGraph", subject);
            var resultSet = QueryEndpoint.QueryWithResultSet(cmdString.ToString());
            return resultSet.First().Value("o").ToString();
        }

        public Dictionary<string,string> GetOrganizationToIkzMap()
        {
            SparqlParameterizedString cmdString = new SparqlParameterizedString
            {
                CommandText = "SELECT ?s SUBSTR( ?o, 5 ) AS ?o WHERE {{ GRAPH <https://ror.org/04xfq0f34> { ?s <http://www.w3.org/ns/org#identifier> ?o .FILTER strStarts( ?o, \"ikz:\" )}}}"
            };
            var resultSet = QueryEndpoint.QueryWithResultSet(cmdString.ToString());
            var dict = new Dictionary<string,string>();
            foreach (SparqlResult r in resultSet)
            {
                dict.Add(r.Value("s").ToString(), r.Value("o").ToString());
            }
            return dict;
        }



        // Find the orgnization by the entityId or by the user identifier.
        public string GetOrganization(string entityId, string identifier)
        {
            string organization = null;

            if (entityId != null)
            {
                organization = GetOrgnizationWithEntityId(entityId);
            }

            if (organization != null)
            {
                return organization;
            }

            if (identifier != null)
            {
                return GetOrgnizationWithIdentifier(identifier);
            }

            return null;
        }

        // Creates graph if needed (silent disables the "error" for existing graph).
        public void EnsureGraph(string graphIri)
        {
            var commandString = new SparqlParameterizedString
            {
                CommandText = "CREATE SILENT GRAPH @graphIri"
            };
            commandString.SetUri("graphIri", new Uri(graphIri));
            QueryEndpoint.QueryRaw(commandString.ToString());
        }

        // Find organization by entityId.
        public string GetOrgnizationWithEntityId(string entityId)
        {
            var commandString = new SparqlParameterizedString
            {
                CommandText = @"SELECT DISTINCT ?organization
                                WHERE
                                {
                                    ?organization org:identifier @entityId .
                                }"
            };
            commandString.Namespaces.AddNamespace("org", OrgPrefixUrl);
            commandString.SetLiteral("entityId", entityId);
            var resultSet = QueryEndpoint.QueryWithResultSet(commandString.ToString());

            if (resultSet.Count != 1)
            {
                return null;
            }

            return resultSet.First().Value("organization").ToString();
        }

        // Find organization by user identifier.
        public string GetOrgnizationWithIdentifier(string identifier)
        {
            var commandString = new SparqlParameterizedString
            {
                CommandText = @"SELECT DISTINCT ?organization
                                WHERE 
                                {
                                    ?organization org:hasUnit ?subOrganization .
                                    ?organization a org:FormalOrganization .
                                	?nodeId org:organization ?subOrganization .
                                    ?nodeId org:member ?organizationMember .
                                	?organizationMember  foaf:openId @identifier .
                                }"
            };
            commandString.Namespaces.AddNamespace("foaf", FoafPrefixUrl);
            commandString.Namespaces.AddNamespace("org", OrgPrefixUrl);
            commandString.SetLiteral("identifier", identifier);

            var resultSet = QueryEndpoint.QueryWithResultSet(commandString.ToString());

            if (resultSet.Count != 1)
            {
                return null;
            }

            return resultSet.First().Value("organization").ToString();
        }

        // Add the membership block to the user graph.
        public void AddMemebershipData(string userGraph, string organization)
        {
            var commandString = new SparqlParameterizedString
            {
                CommandText = @"INSERT
                                {
                                    GRAPH @userGraph
                                    {
                                        [
                                            a org:Membership ;
                                            org:member @member ;
                                            org:organization @organization ;
                                        ]
                                    }
                                }"
            };

            commandString.Namespaces.AddNamespace("org", OrgPrefixUrl);
            commandString.SetUri("userGraph", new Uri(userGraph));
            commandString.SetUri("member", new Uri(userGraph));
            commandString.SetUri("organization", new Uri(organization));
            QueryEndpoint.QueryRaw(commandString.ToString());
        }

        // Remove the membership block to the user graph.
        public void RemoveMembershipData(string userGraph, string organization)
        {
            var commandString = new SparqlParameterizedString
            {
                CommandText = @"DELETE
                                {
                                    GRAPH @userGraph
                                    {
                                        ?s ?p0 ?o0 .
                                    }
                                }

                                USING @userGraph WHERE
                                {
                                    ?s ?p0 ?o0 .
                                    ?s a org:Membership .
                                    ?s org:member @member .
                                    ?s org:organization @organization .
                                }"
            };

            commandString.Namespaces.AddNamespace("org", OrgPrefixUrl);
            commandString.SetUri("userGraph", new Uri(userGraph));
            commandString.SetUri("member", new Uri(userGraph));
            commandString.SetUri("organization", new Uri(organization));
            QueryEndpoint.QueryRaw(commandString.ToString());
        }

        /// <summary>
        /// Checks if the given user is a member of the RWTH 
        /// </summary>
        /// <param name="user">User object</param>
        /// <returns>A dictonary with resource type name and default quota.</returns>
        public Dictionary<string, int> GetQuotaDefault(string userId)
        {
            var userGraphName = $"{UserUrlPrefix}/{userId}";

            var commandString = new SparqlParameterizedString
            {
                CommandText = @"SELECT ?resourceType MAX(?dQuota) as ?defaultQuota
                                WHERE
                                {
                                    ?bNode2 coscineresource:type ?resourceType .
                                    ?bNode2 coscineresource:defaultQuota ?dQuota .
                                    ?organization coscineresource:typeSpecification ?bNode2 .
                                    ?bNode1 org:organization ?organization .
                                
                                    GRAPH @userGraph
                                    {
                                        ?bNode1 org:member @userGraph .
                                    }
                                }"
            };

            commandString.Namespaces.AddNamespace("org", OrgPrefixUrl);
            commandString.Namespaces.AddNamespace("coscineresource", new Uri("https://purl.org/coscine/terms/resource#"));
            commandString.SetUri("userGraph", new Uri(userGraphName));
            var resultSet = QueryEndpoint.QueryWithResultSet(commandString.ToString());
            return resultSet.ToDictionary(
                keySelector: x =>
                {
                    var resourceType = x.Value("resourceType").ToString();
                    return resourceType[(resourceType.LastIndexOf("#") + 1)..];
                },
                elementSelector: x => int.Parse(x.Value("defaultQuota").ToString())
                );
        }
    }
}
